@@ -3,34 +3,69 @@ import base64
 import json
 
 
-class ACLDisabled(Exception):
+class ConsulException(Exception):
     pass
 
 
-class ACLPermissionDenied(Exception):
+class ACLDisabled(ConsulException):
     pass
 
 
-class Timeout(Exception):
+class ACLPermissionDenied(ConsulException):
+    pass
+
+
+class Timeout(ConsulException):
     pass
 
 
 Response = collections.namedtuple('Response', ['code', 'headers', 'body'])
 
 
+def callback(callback=None, is_200=False, is_indexed=False):
+    def cb(response):
+        if response.code == 500:
+            raise ConsulException(response.body)
+        if is_200:
+            return response.code == 200
+        if is_indexed:
+            return (
+                response.headers['X-Consul-Index'], json.loads(response.body))
+        if callback:
+            return callback(response)
+        return response
+    return cb
+
+
 class Consul(object):
-    def __init__(self, host='127.0.0.1', port=8500, token=None):
+    def __init__(
+            self,
+            host='127.0.0.1',
+            port=8500,
+            token=None,
+            consistency='default'):
         """
         *token* is an optional `ACL token`_. If supplied it will be used by
         default for all requests made with this client session. It's still
         possible to override this token by passing a token explicitly for a
         request.
+
+        *consistency* sets the consistency mode to use by default for all reads
+        that support the consistency option. It's still possible to override
+        this by passing explicitly for a given request. *consistency* can be
+        either 'default', 'consistent' or 'stale'.
         """
+
+        # TODO: Session, Event, Status
+
         self.http = self.connect(host, port)
         self.token = token
+        assert consistency in ('default', 'consistent', 'stale')
+        self.consistency = consistency
 
         self.kv = Consul.KV(self)
         self.agent = Consul.Agent(self)
+        self.catalog = Consul.Catalog(self)
         self.health = Consul.Health(self)
         self.acl = Consul.ACL(self)
 
@@ -43,7 +78,13 @@ class Consul(object):
         def __init__(self, agent):
             self.agent = agent
 
-        def get(self, key, index=None, recurse=False, token=None):
+        def get(
+                self,
+                key,
+                index=None,
+                recurse=False,
+                token=None,
+                consistency=None):
             """
             Returns a tuple of (*index*, *value[s]*)
 
@@ -77,6 +118,9 @@ class Consul(object):
             token = token or self.agent.token
             if token:
                 params['token'] = token
+            consistency = consistency or self.agent.consistency
+            if consistency in ('consistent', 'stale'):
+                params[consistency] = '1'
 
             def callback(response):
                 if response.code == 404:
@@ -164,6 +208,8 @@ class Consul(object):
         takes on the burden of registering with the Catalog and performing
         anti-entropy to recover from outages.
         """
+        # TODO: checks, members, join, force-leave
+        # TODO: Check. register, deregister, pass (move), warn, fail
         def __init__(self, agent):
             self.agent = agent
             self.service = Consul.Agent.Service(agent)
@@ -239,7 +285,290 @@ class Consul(object):
                     lambda x: x.code == 200,
                     '/v1/agent/service/deregister/%s' % service_id)
 
+    class Catalog(object):
+        def __init__(self, agent):
+            self.agent = agent
+
+        def register(self, node, address, dc=None, service=None, check=None):
+            """
+            A low level mechanism for directly registering or updating entries
+            in the catalog. It is usually recommended to use
+            agent.service.register and agent.check.register, as they are
+            simpler and perform anti-entropy.
+
+            *node* is the name of the node to register.
+
+            *address* is the ip of the node.
+
+            *dc* is the datacenter of the node and defaults to this agents
+            datacenter.
+
+            *service* is an optional service to register. if supplied this is a
+            dict::
+
+                {
+                    "Service": "redis",
+                    "ID": "redis1",
+                    "Tags": [
+                        "master",
+                        "v1"
+                    ],
+                    "Port": 8000
+                }
+
+            where
+
+                *Service* is required and is the name of the service
+
+                *ID* is optional, and will be set to *Service* if not provided.
+                Note *ID* must be unique for the given *node*.
+
+                *Tags* and *Port* are optional.
+
+            *check* is an optional check to register. if supplied this is a
+            dict::
+
+                {
+                    "Node": "foobar",
+                    "CheckID": "service:redis1",
+                    "Name": "Redis health check",
+                    "Notes": "Script based health check",
+                    "Status": "passing",
+                    "ServiceID": "redis1"
+                }
+
+            This manipulates the health check entry, but does not setup a
+            script or TTL to actually update the status. The full documentation
+            is `here <https://consul.io/docs/agent/http.html#catalog>`_.
+
+            Returns *True* on success.
+            """
+            data = {'node': node, 'address': address}
+            if dc:
+                data['datacenter'] = dc
+            if service:
+                data['service'] = service
+            if check:
+                data['check'] = check
+            return self.agent.http.put(
+                callback(is_200=True),
+                '/v1/catalog/register', data=json.dumps(data))
+
+        def deregister(self, node, dc=None, service_id=None, check_id=None):
+            """
+            A low level mechanism for directly removing entries in the catalog.
+            It is usually recommended to use the agent APIs, as they are
+            simpler and perform anti-entropy.
+
+            *node* and *dc* specify which node on which datacenter to remove.
+            If *service_id* and *check_id* are not provided, all associated
+            services and checks are deleted. Otherwise only one of *service_id*
+            and *check_id* should be provided and only that service or check
+            will be removed.
+
+            Returns *True* on success.
+            """
+            assert not (service_id and check_id)
+            data = {'node': node}
+            if dc:
+                data['datacenter'] = dc
+            if service_id:
+                data['serviceid'] = service_id
+            if check_id:
+                data['checkid'] = check_id
+            return self.agent.http.put(
+                callback(is_200=True),
+                '/v1/catalog/deregister', data=json.dumps(data))
+
+        def datacenters(self):
+            """
+            Returns all the datacenters that are known by the Consul server.
+            """
+            return self.agent.http.get(
+                lambda x: json.loads(x.body), '/v1/catalog/datacenters')
+
+        def nodes(self, dc=None, index=None, consistency=None):
+            """
+            Returns a tuple of (*index*, *nodes*) of all nodes known
+            about in the *dc* datacenter. *dc* defaults to the current
+            datacenter of this agent.
+
+            *index* is the current Consul index, suitable for making subsequent
+            calls to wait for changes since this query was last run.
+
+            *consistency* can be either 'default', 'consistent' or 'stale'. if
+            not specified *consistency* will the consistency level this client
+            was configured with.
+
+            The response looks like this::
+
+                (index, [
+                    {
+                        "Node": "baz",
+                        "Address": "10.1.10.11"
+                    },
+                    {
+                        "Node": "foobar",
+                        "Address": "10.1.10.12"
+                    }
+                ])
+            """
+            params = {}
+            if dc:
+                params['dc'] = dc
+            if index:
+                params['index'] = index
+            consistency = consistency or self.agent.consistency
+            if consistency in ('consistent', 'stale'):
+                params[consistency] = '1'
+            return self.agent.http.get(
+                callback(is_indexed=True), '/v1/catalog/nodes', params=params)
+
+        def services(self, dc=None, index=None, consistency=None):
+            """
+            Returns a tuple of (*index*, *services*) of all services known
+            about in the *dc* datacenter. *dc* defaults to the current
+            datacenter of this agent.
+
+            *index* is the current Consul index, suitable for making subsequent
+            calls to wait for changes since this query was last run.
+
+            *consistency* can be either 'default', 'consistent' or 'stale'. if
+            not specified *consistency* will the consistency level this client
+            was configured with.
+
+            The response looks like this::
+
+                (index, {
+                    "consul": [],
+                    "redis": [],
+                    "postgresql": [
+                        "master",
+                        "slave"
+                    ]
+                })
+
+            The main keys are the service names and the list provides all the
+            known tags for a given service.
+            """
+            params = {}
+            if dc:
+                params['dc'] = dc
+            if index:
+                params['index'] = index
+            consistency = consistency or self.agent.consistency
+            if consistency in ('consistent', 'stale'):
+                params[consistency] = '1'
+            return self.agent.http.get(
+                callback(is_indexed=True),
+                '/v1/catalog/services', params=params)
+
+        def node(self, node, dc=None, index=None, consistency=None):
+            """
+            Returns a tuple of (*index*, *services*) of all services provided
+            by *node*.
+
+            *dc* is the datacenter of the node and defaults to this agents
+            datacenter.
+
+            *index* is the current Consul index, suitable for making subsequent
+            calls to wait for changes since this query was last run.
+
+            *consistency* can be either 'default', 'consistent' or 'stale'. if
+            not specified *consistency* will the consistency level this client
+            was configured with.
+
+            The response looks like this::
+
+                (index, {
+                    "Node": {
+                        "Node": "foobar",
+                        "Address": "10.1.10.12"
+                    },
+                    "Services": {
+                        "consul": {
+                            "ID": "consul",
+                            "Service": "consul",
+                            "Tags": null,
+                            "Port": 8300
+                        },
+                        "redis": {
+                            "ID": "redis",
+                            "Service": "redis",
+                            "Tags": [
+                                "v1"
+                            ],
+                            "Port": 8000
+                        }
+                    }
+                })
+            """
+            params = {}
+            if dc:
+                params['dc'] = dc
+            if index:
+                params['index'] = index
+            consistency = consistency or self.agent.consistency
+            if consistency in ('consistent', 'stale'):
+                params[consistency] = '1'
+            return self.agent.http.get(
+                callback(is_indexed=True),
+                '/v1/catalog/node/%s' % node, params=params)
+
+        def service(
+                self,
+                service,
+                dc=None,
+                tag=None,
+                index=None,
+                consistency=None):
+            """
+            Returns a tuple of (*index*, *nodes*) of the nodes providing
+            *service* in the *dc* datacenter. *dc* defaults to the current
+            datacenter of this agent.
+
+            If *tag* is provided, the list of nodes returned will be filtered
+            by that tag.
+
+            *index* is the current Consul index, suitable for making subsequent
+            calls to wait for changes since this query was last run.
+
+            *consistency* can be either 'default', 'consistent' or 'stale'. if
+            not specified *consistency* will the consistency level this client
+            was configured with.
+
+            The response looks like this::
+
+                (index, [
+                    {
+                        "Node": "foobar",
+                        "Address": "10.1.10.12",
+                        "ServiceID": "redis",
+                        "ServiceName": "redis",
+                        "ServiceTags": null,
+                        "ServicePort": 8000
+                    }
+                ])
+            """
+            params = {}
+            if dc:
+                params['dc'] = dc
+            if tag:
+                params['tag'] = tag
+            if index:
+                params['index'] = index
+            consistency = consistency or self.agent.consistency
+            if consistency in ('consistent', 'stale'):
+                params[consistency] = '1'
+            return self.agent.http.get(
+                callback(is_indexed=True),
+                '/v1/catalog/service/%s' % service, params=params)
+
     class Health(object):
+        # TODO: All of the health endpoints support blocking queries and all
+        # consistency modes
+        # TODO: Check ttl_pass belongs under Agent
+        # TODO: still need to add node, checks and state endpoints
         def __init__(self, agent):
             self.agent = agent
             self.check = Consul.Health.Check(agent)
