@@ -32,6 +32,10 @@ class HTTPClient(object):
         uri = self.uri(path, params)
         return callback(self.response(requests.put(uri, data=data)))
 
+    def delete(self, callback, path, params=None):
+        uri = self.uri(path, params)
+        return callback(self.response(requests.delete(uri, params=params)))
+
 
 class Consul(base.Consul):
     def connect(self, host, port):
